mod daemon;
mod listenbrainz;
mod models;
mod playlist;
mod subsonic;
mod track;

use crate::daemon::update_playlists_daemon;
use crate::models::{PartialTrackedPlaylist, TrackedPlaylist};
use crate::subsonic::SubsonicBuilder;
use axum::extract::State;
use axum::routing::{get, put};
use axum::{extract, Json, Router};
use config::Config;
use serde::Serialize;
use sqlx::postgres::PgPool;
use sqlx::Error;
use std::sync::OnceLock;

pub static CONFIG: OnceLock<Config> = OnceLock::new();

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    CONFIG
        .set(
            Config::builder()
                .add_source(config::File::with_name("/etc/playlistd").required(false))
                .add_source(config::Environment::default().separator("_"))
                .build()
                .unwrap(),
        )
        .unwrap();

    let database = PgPool::connect(
        CONFIG
            .get()
            .unwrap()
            .get::<String>("database.url")?
            .as_str(),
    )
    .await?;
    sqlx::migrate!().run(&database).await?;

    let media_client = SubsonicBuilder::new()
        .base(CONFIG.get().unwrap().get::<String>("subsonic.base")?)
        .username(CONFIG.get().unwrap().get::<String>("subsonic.username")?)
        .password(CONFIG.get().unwrap().get::<String>("subsonic.password")?)
        .build()?;

    tokio::spawn(async move {
        update_playlists_daemon(&media_client).await.unwrap();
    });

    let app = Router::new()
        .route("/playlists", get(all_playlists))
        .route("/playlists", put(create_tracking_playlist))
        .with_state(database);

    let listener = tokio::net::TcpListener::bind("localhost:3000").await?;
    axum::serve(listener, app).await?;

    Ok(())
}

#[derive(Serialize)]
struct JsonError {
    error: String,
}

impl From<Error> for JsonError {
    fn from(value: Error) -> Self {
        Self {
            error: value.to_string(),
        }
    }
}

#[derive(Serialize)]
enum Response<T> {
    Success(T),
    Error(JsonError),
}

async fn create_tracking_playlist(
    State(pool): State<PgPool>,
    extract::Json(partial): extract::Json<PartialTrackedPlaylist>,
) -> Json<Response<TrackedPlaylist>> {
    match partial.record(&pool).await {
        Ok(playlist) => Json(Response::Success(playlist)),
        Err(e) => Json(Response::Error(e.into())),
    }
}

async fn all_playlists(State(pool): State<PgPool>) -> Json<Vec<TrackedPlaylist>> {
    match TrackedPlaylist::all(&pool).await {
        Ok(playlists) => Json(playlists),
        Err(_) => Json(vec![]),
    }
}
